#include <postgres.h>
#include <fmgr.h>
#include <funcapi.h>
#include <executor/spi.h>
#include <commands/trigger.h>
#include <utils/lsyscache.h>
#define GET_TEXT(cp) DatumGetTextP(DirectFunctionCall1(textin, CStringGetDatum(cp)))
#define GET_STR(textp) DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(textp)))
#define XML_ROWTYPE "get_xml_row_type"
#define HEAD "<?xml version = '1.0'?>\n<rowset>"
#define ROWHEAD "<row>\n"
#define ROWTAIL "</row>"
#define TAIL   "</rowset>"
#define ROWMASK "  <%s>%s</%s>"
#define HEADTAILLEN 2
#define NULO ""

extern Datum get_xml(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(get_xml);

Datum get_xml(PG_FUNCTION_ARGS)
{
static int filas;
char *comando = GET_STR (PG_GETARG_TEXT_P(0));
FuncCallContext *funcctx;
int call_cntr;
int max_calls,largo;
TupleTableSlot *slot;
AttInMetadata *attinmeta;
MemoryContext oldcontext;
static int header,exit=0,last=0,j=0,i=1;
int ret,rowfree=0;
char *row,*clave,*valor;
char *rowarr[2];
static TupleDesc tupdesc;
TupleDesc newtupdesc;
static SPITupleTable *tuptable;

        /* First call */
        if (SRF_IS_FIRSTCALL())
        {
          if ((ret=SPI_connect()) < 0)
              {
              elog(ERROR, "get_xml (%d) - can't connect SPI",ret);
              PG_RETURN_NULL();
              }

          if ((ret=SPI_exec(comando,0)) != SPI_OK_SELECT)
            {
            elog(ERROR, "get_xml (%d) - SPI exec failed : query = %s",ret,comando);
            PG_RETURN_NULL();
            }

          filas=SPI_processed; 
          tupdesc = SPI_tuptable->tupdesc;
          tuptable = SPI_tuptable;
           
          funcctx = SRF_FIRSTCALL_INIT();
          oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
          funcctx->max_calls = filas+ HEADTAILLEN; 
          newtupdesc = RelationNameGetTupleDesc(XML_ROWTYPE);

          slot = TupleDescGetSlot(newtupdesc);
          funcctx->slot = slot;
         
          attinmeta = TupleDescGetAttInMetadata(newtupdesc);
          funcctx->attinmeta = attinmeta;

          MemoryContextSwitchTo(oldcontext);
          header=1;
          exit=0;
          last=0;
          j=0;
         }
 
         funcctx = SRF_PERCALL_SETUP();
         call_cntr = funcctx->call_cntr;
         max_calls = funcctx->max_calls;
         slot = funcctx->slot;
         attinmeta = funcctx->attinmeta;

       if (header) {
          row=HEAD;
          rowfree=0;
          header=0;
       } else {
        if (j < filas) {
            HeapTuple tuple = tuptable->vals[j];
            if (i <= tupdesc->natts) {
                clave=SPI_fname(tupdesc, i);
                if (clave==NULL) clave=NULO; 
                valor=SPI_getvalue(tuple, tupdesc, i);
                if (valor==NULL) valor=NULO;
                largo=(strlen(clave)+strlen(clave)+strlen(valor)+strlen(ROWMASK)+1);
                if (i==1)  
                     largo += strlen(ROWHEAD);

                row=palloc(largo);
                rowfree=1;
                
                if (i==1) {
                strcpy(row,ROWHEAD);
                sprintf(row+strlen(ROWHEAD),ROWMASK,clave,valor,clave); 
                } else { 
                sprintf(row,ROWMASK,clave,valor,clave);
                } 
                i++;
                if (strlen(clave))
                        pfree(clave);
                if (strlen(valor))
                        pfree(valor);  
            } else {
               j++; i=1; 
               row=ROWTAIL; 
            }
         } else {
             /* Tail */
             row=TAIL;
             rowfree=0;
             last=1;
         } 
        }

        if (!exit) 
        {
        HeapTuple       tuple;
        Datum           result;

        if (last) exit=1;
         
        /* build the tuple */
        rowarr[0]=row;
        rowarr[1]=NULL;
        tuple = BuildTupleFromCStrings(attinmeta, rowarr);
        /* Free row */
        if (rowfree) pfree(row);
        /* make the tuple into a datum */
        result = TupleGetDatum(slot, tuple);
        SRF_RETURN_NEXT(funcctx, result);
        }
        else
        /* do when there is no more left */
        {
                SPI_finish();
                SRF_RETURN_DONE(funcctx);
        }
}
